/*
  FIR image filter
  It performs a smooth operation over the input image

  Note that only the method doFIR will be used (executed) in the GUI interface
  that appears with: java DrawImage

  Thus, to evaluate different implementations of doFIR you have to change the
  method names.

 */

public class Filter {
    // input/output image size (fixed because of the example used in the GUI)
    final static int sizeX = 350;
    final static int sizeY = 350;

	// to count the number of loads and stores from/to external memories
    static int numExtLoads = 0;
    static int numExtStores = 0;

	// to count the number of loads and stores from/to internal buffers
    static int numIntLoads = 0;
    static int numIntStores = 0;

    /* The FIR filter (Image Smooth).
    IN represents the input image in gray scale
    The method returns the output image in gray scale.
     */
    final public static short[] doFIR(short[] IN) {

        short[] K = {1, 2, 1,
            2, 4, 2,
            1, 2, 1
        };

        short[] OUT = new short[sizeY * sizeX];

        for (int row = 0; row < sizeY - 3 + 1; row++) {
            for (int col = 0; col < sizeX - 3 + 1; col++) {
                int sumval = 0;

                for (int wrow = 0; wrow < 3; wrow++) {
                    for (int wcol = 0; wcol < 3; wcol++) {
                        sumval += ((IN[(row + wrow) * 350 + (col + wcol)]) * K[wrow * 3 + wcol]);

                        numExtLoads += 2; // just to count
                    }
                }

                sumval = sumval / 16;
                OUT[row * 350 + col] = (short) sumval;

                numExtStores++;  // just to count
            }
        }

        // report number of external loads and stores
        System.out.println("#External Loads: " + numExtLoads + " #External Stores: " + numExtStores);
        return OUT;
    }
    // the same implementation, but using two-dimensional arrays
    // to store images
    final public static short[] doFIR2(short[] IN) {

        short[] OUT = new short[sizeY * sizeX];

        short[][] IN1 = new short[sizeY][sizeX];
        short[][] OUT1 = new short[sizeY][sizeX];

        // needed as the original code considers images stored as unidimension
        // arrays
        for (int i = 0; i < sizeY; i++) {
            for (int j = 0; j < sizeX; j++) {
                IN1[i][j] = IN[i * sizeX + j];
            }
        }
        short[][] K = {{1, 2, 1},
            {2, 4, 2},
            {1, 2, 1}
        };

        for (int j = 0; j < sizeY - 2; j++) {
            for (int i = 0; i < sizeX - 2; i++) {
                int sum = 0;
                for (int r = 0; r < 3; r++) {
                    for (int c = 0; c < 3; c++) {
                        sum += IN1[j + r][i + c] * K[r][c];
                    }
                }
                sum = sum / 16;
                OUT1[j + 1][i + 1] = (short) sum;
            }
        }

        // needed as the original code considers images stored as unidimension
        // arrays
        for (int i = 0; i < sizeY; i++) {
            for (int j = 0; j < sizeX; j++) {
                OUT[i * sizeX + j] = OUT1[i][j];
            }
        }
        return OUT;
    }

    // an implementation with the two inner loops fully unrolled
    final public static short[] doFIR3(short[] IN) {

        short[] OUT = new short[sizeY * sizeX];

        short[][] IN1 = new short[sizeY][sizeX];
        short[][] OUT1 = new short[sizeY][sizeX];

        for (int i = 0; i < sizeY; i++) {
            for (int j = 0; j < sizeX; j++) {
                IN1[i][j] = IN[i * sizeX + j];
            }
        }
        for (int j = 0; j < sizeY - 2; j++) {
            for (int i = 0; i < sizeX - 2; i++) {
                int sum = IN1[j][i];
                sum += IN1[j][i + 1] * 2;
                sum += IN1[j][i + 2];
                sum += IN1[j + 1][i] * 2;
                sum += IN1[j + 1][i + 1] * 4;
                sum += IN1[j + 1][i + 2] * 2;
                sum += IN1[j + 2][i];
                sum += IN1[j + 2][i + 1] * 2;
                sum += IN1[j + 2][i + 2];

                numExtLoads += 9; // just to monitor

                sum = sum / 16;


                numExtStores += 1; // just to monitor
                OUT1[j + 1][i + 1] = (short) sum;
            }
        }


        // report number of external loads and stores
        System.out.println("#External Loads: " + numExtLoads + " #External Stores: " + numExtStores);


        for (int i = 0; i < sizeY; i++) {
            for (int j = 0; j < sizeX; j++) {
                OUT[i * sizeX + j] = OUT1[i][j];
            }
        }
        return OUT;
    }

    // an example of an implementation using data reuse
    final public static short[] doFIRa(short[] IN) {

        short[] OUT = new short[sizeY * sizeX];

        short[][] IN1 = new short[sizeY][sizeX];
        short[][] OUT1 = new short[sizeY][sizeX];

        for (int i = 0; i < sizeY; i++) {
            for (int j = 0; j < sizeX; j++) {
                IN1[i][j] = IN[i * sizeX + j];
            }
        }
        short[] l1 = new short[sizeX];
        short[] l2 = new short[sizeX];
        short[] l3 = new short[sizeX];
        int r11, r12, r13, r21, r22, r23, r31, r32, r33;


        for (int j = 0; j < sizeY - 2; j++) {
            for (int i = 0; i < sizeX - 2; i++) {
                if (j == 0) {
                    l1[i] = IN1[0][i];
                    l1[i + 1] = IN1[0][i + 1];
                    l1[i + 2] = IN1[0][i + 2];
                    l2[i] = IN1[1][i];
                    l2[i + 1] = IN1[1][i + 1];
                    l2[i + 2] = IN1[1][i + 2];
                    l3[i] = IN1[2][i];
                    l3[i + 1] = IN1[2][i + 1];
                    l3[i + 2] = IN1[2][i + 2];

                    numExtLoads += 9; // just to count
                    numIntStores += 9; // just to count
                }

                if (j > 0) {
                    l1 = l2;
                    l2 = l3;
                    l3[i] = IN1[j + 2][i];
                    l3[i + 1] = IN1[j + 2][i + 1];
                    l3[i + 2] = IN1[j + 2][i + 2];
                    numExtLoads += 3; // just to count
                    numIntStores += 3; // just to count
                }


                r11 = l1[i];
                r12 = l1[i + 1];
                r13 = l1[i + 2];
                r21 = l2[i];
                r22 = l2[i + 1];
                r23 = l2[i + 2];
                r31 = l3[i];
                r32 = l3[i + 1];
                r33 = l3[i + 2];

                numIntLoads += 9; // just to count

                int sum;
                sum = calc(r11, r12, r13, r21, r22, r23, r31, r32, r33);

                numExtStores += 1; // just to monitor
                OUT1[j + 1][i + 1] = (short) sum;



            }
        }

        // report number of external loads and stores
        System.out.println("#External Loads: " + numExtLoads + " #External Stores: " + numExtStores);

        // report number of internal loads and stores
        System.out.println("#Internal Loads: " + numIntLoads + " #Internal Stores: " + numIntStores);

        for (int i = 0; i < sizeY; i++) {
            for (int j = 0; j < sizeX; j++) {
                OUT[i * sizeX + j] = OUT1[i][j];
            }
        }
        return OUT;
    }

    public final static int calc(int l11, int l12, int l13, int l21,
            int l22, int l23, int l31, int l32, int l33) {
        int sum = l11;
        sum += l12 * 2;
        sum += l13;
        sum += l21 * 2;
        sum += l22 * 4;
        sum += l23 * 2;
        sum += l31;
        sum += l32 * 2;
        sum += l33;
        sum = sum / 16;
        return sum;
    }

	// an example of an implementation using more aggressive data reuse
    final public static short[] doFIRb(short[] IN) {

        short[] OUT = new short[sizeY * sizeX];

        short[][] IN1 = new short[sizeY][sizeX];
        short[][] OUT1 = new short[sizeY][sizeX];

        for (int i = 0; i < sizeY; i++) {
            for (int j = 0; j < sizeX; j++) {
                IN1[i][j] = IN[i * sizeX + j];
            }
        }
        short[] l1 = new short[sizeX];
        short[] l2 = new short[sizeX];
        short[] l3 = new short[sizeX];
        int r11 = 0, r12 = 0, r13 = 0, r21 = 0, r22 = 0, r23 = 0, r31 = 0, r32 = 0, r33 = 0;


        for (int j = 0; j < sizeY - 2; j++) {
            for (int i = 0; i < sizeX - 2; i++) {
                if (j == 0) {
                    if (i == 0) {
                        r11 = IN1[0][0];
                        r12 = IN1[0][1];
                        r13 = IN1[0][2];
                        r21 = IN1[1][0];
                        r22 = IN1[1][1];
                        r23 = IN1[1][2];
                        r31 = IN1[2][0];
                        r32 = IN1[2][1];
                        r33 = IN1[2][2];
                        l1[0] = (short) r11;
                        l1[1] = (short) r12;
                        l1[2] = (short) r13;
                        l2[0] = (short) r21;
                        l2[1] = (short) r22;
                        l2[2] = (short) r23;
                        l3[0] = (short) r31;
                        l3[1] = (short) r32;
                        l3[2] = (short) r33;

                        numExtLoads += 9; // just to count
                        numIntStores += 9; // just to count
                    } else {
                        r11 = r12;
                        r12 = r13;
                        r13 = IN1[0][i + 2];
                        r21 = r22;
                        r22 = r23;
                        r23 = IN1[1][i + 2];
                        r31 = r32;
                        r32 = r33;
                        r33 = IN1[2][i + 2];
                        l1[i + 2] = (short) r13;
                        l2[i + 2] = (short) r23;
                        l3[i + 2] = (short) r33;

                        numExtLoads += 3; // just to count
                        numIntStores += 3; // just to count
                    }

                } else {
                    if (i == 0) {

                        l1 = l2;
                        l2 = l3;
                        r31 = IN1[j + 2][0];
                        r32 = IN1[j + 2][1];
                        r33 = IN1[j + 2][2];
                        l3[0] = (short) r31;
                        l3[1] = (short) r32;
                        l3[2] = (short) r33;

                        r11 = l1[0];
                        r12 = l1[1];
                        r13 = l1[2];
                        r21 = l2[0];
                        r22 = l2[1];
                        r23 = l2[2];

                        numExtLoads += 3; // just to count
                        numIntStores += 3; // just to count
                        numIntLoads += 6; // just to count
                    } else {
                        r11 = r12;
                        r12 = r13;
                        r13 = l1[i + 2];
                        r21 = r22;
                        r22 = r23;
                        r23 = l2[i + 2];
                        r31 = r32;
                        r32 = r33;
                        r33 = IN1[j + 2][i + 2];

                        l3[i + 2] = (short) r33;

                        numExtLoads += 1; // just to count
                        numIntStores += 1; // just to count
                        numIntLoads += 2; // just to count
                    }
                }


                int sum;
                sum = calc(r11, r12, r13, r21, r22, r23, r31, r32, r33);

                numExtStores += 1; // just to count
                OUT1[j + 1][i + 1] = (short) sum;



            }
        }

        // report number of external loads and stores
        System.out.println("#External Loads: " + numExtLoads + " #External Stores: " + numExtStores);

        // report number of internal loads and stores
        System.out.println("#Internal Loads: " + numIntLoads + " #Internal Stores: " + numIntStores);

        for (int i = 0; i < sizeY; i++) {
            for (int j = 0; j < sizeX; j++) {
                OUT[i * sizeX + j] = OUT1[i][j];
            }
        }
        return OUT;
    }
}
